import time
import botocore
import boto3
import pyarrow as pa
import pyarrow.parquet as pq
from datetime import datetime

from s3transfer import S3UploadFailedError

from airflow_commons.internal.util.time_utils import get_interval_duration
from airflow_commons.logger import get_logger
from s3fs import S3FileSystem


class S3FileSystemOperator(object):
    DEFAULT_RETRY_COUNT = 3
    DEFAULT_CONTENT_TYPE = "str"

    def __init__(self, key: str = None, secret: str = None):
        """
        Initializes s3FileSystemOperator instance with given credentials.
        If no credentials were supplied, uses environment variables.

        :param key:
        :param secret:
        """
        self.s3_file_system = S3FileSystem(key=key, secret=secret)
        self.logger = get_logger("S3FileSystemOperator")

    def write_into_s3_file(
        self,
        bucket_name: str,
        file_name: str,
        data,
        content_type: str = DEFAULT_CONTENT_TYPE,
        retry_count: int = DEFAULT_RETRY_COUNT,
    ):
        """
        Writes the given string data into the specified file in the specified bucket.
        If file does not exist creates one, if exists overrides it.

        :param bucket_name: Name of the bucket that the target file is stored
        :param file_name: Name of the file that will be overridden
        :param data: An object contains the content of the file
        :param content_type: A string that describes the type of the contents of the file, can take values "str", "pyarrow.lib.Table" and "bytes"
        :param retry_count: retry count for S3 upload equals to three on default
        """
        writing_start = datetime.now()
        total_upload_tries = 0
        if content_type == "pyarrow.lib.Table":
            output_file = f"s3://{bucket_name}/{file_name}"
            pq.write_table(
                table=data, where=output_file, filesystem=self.s3_file_system
            )
        else:
            write_mode = "w"
            if content_type == "bytes":
                write_mode = "wb"
            while total_upload_tries <= retry_count:
                with self.s3_file_system.open(
                    bucket_name + "/" + file_name, write_mode
                ) as f:
                    try:
                        f.write(data)
                        break
                    except botocore.exceptions.NoCredentialsError as e:
                        total_upload_tries = total_upload_tries + 1
                        if total_upload_tries == retry_count:
                            self.logger.error(
                                f"Writing into {bucket_name}/{file_name} failed because of missing credentials, traceback {e}"
                            )
                            raise e
                        time.sleep(1)
        writing_end = datetime.now()
        self.logger.info(
            f"Writing finished in {get_interval_duration(writing_start, writing_end)} seconds"
        )

    def write_to_s3_with_parquet(
        self, bucket_name: str, container_name: str, table: pa.Table
    ):
        """
        Writes the given string data into the specified file in the specified bucket.
        :param bucket_name: Name of the bucket that the target file is stored
        :param container_name: Name of the container that will be overridden
        :param table: Table that will be written to the dataset whose filepath created by bucket_name and container_name
        """
        output_file = f"s3://{bucket_name}/{container_name}"
        pq.write_to_dataset(
            table=table, root_path=output_file, filesystem=self.s3_file_system
        )

    def move_s3_file(
        self,
        source_path: str,
        destination_path: str,
        retry_count: int = DEFAULT_RETRY_COUNT,
    ):
        """
        Move the file in the source path to the destination path. If the file does not exist in the specified
        source path, it raises a FileNotFoundError.
        If the aws key and secret is not given, method uses the environmental variables as credentials.

        :param source_path: The path where the file locates
        :param destination_path: The path where the file will be moved
        :param retry_count: retry count for S3 moving equals to three on default
        """
        moving_start = datetime.now()
        total_move_tries = 0
        while total_move_tries <= retry_count:
            try:
                self.s3_file_system.move(source_path, destination_path)
                break
            except botocore.exceptions.NoCredentialsError as e:
                total_move_tries = total_move_tries + 1
                if total_move_tries == retry_count:
                    self.logger.error(
                        f"Reading from {source_path} to {destination_path} failed because of missing credentials, traceback {e}"
                    )
                    raise e
                time.sleep(1)
        moving_end = datetime.now()
        self.logger.info(
            f"Moving finished in {get_interval_duration(moving_start, moving_end)} seconds"
        )

    def read_from_s3_file(
        self,
        bucket_name: str,
        file_name: str,
        retry_count: int = DEFAULT_RETRY_COUNT,
    ):
        """
        Read data from the specified file in the specified bucket.
        If file does not exists it raises a FileNotFoundError.
        If the aws key and secret is not given, method uses the environmental variables as credentials.

        :param bucket_name: Name of the bucket that the target file is stored
        :param file_name: Name of the file that will be read
        :param retry_count: retry count for S3 reading equals to three on default
        :return: content of the target file in string format
        """
        reading_start = datetime.now()
        total_read_tries = 0
        while total_read_tries <= retry_count:
            with self.s3_file_system.open(bucket_name + "/" + file_name, "rb") as f:
                try:
                    data = f.read()
                    break
                except botocore.exceptions.NoCredentialsError as e:
                    total_read_tries = total_read_tries + 1
                    if total_read_tries == retry_count:
                        self.logger.error(
                            f"Reading from {bucket_name}/{file_name} failed because of missing credentials, traceback {e}"
                        )
                        raise e
                    time.sleep(1)
        reading_end = datetime.now()
        self.logger.info(
            f"Reading finished in {get_interval_duration(reading_start, reading_end)} seconds"
        )
        return data.decode("utf-8")


class S3Operator(object):
    def __init__(self, key: str = None, secret: str = None):
        """
        Initializes S3Operator instance with given credentials.
        If no credentials are supplied, uses environment variables

        :param key: AWS access key id, default is None
        :param secret: AWS secret access key, default is None
        """
        self.client = boto3.client(
            service_name="s3",
            aws_access_key_id=key,
            aws_secret_access_key=secret,
        )
        self.logger = get_logger("S3Operator")

    def upload_file_to_s3_bucket(
        self, path_to_file: str, bucket_name: str, file_name: str
    ):
        """
        Uploads the given file to the given s3 bucket.

        :param path_to_file: Path to file that will be uploaded to s3 bucket.
        :param bucket_name: Name of the bucket that file will be uploaded to.
        :param file_name: Name of the file (key of the file in s3).
        """
        upload_start = datetime.now()
        try:
            self.client.upload_file(path_to_file, bucket_name, file_name)
        except S3UploadFailedError as e:
            self.logger.error(f"Upload to {bucket_name} failed")
            raise e
        upload_end = datetime.now()
        self.logger.info(
            f"Upload finished in {get_interval_duration(upload_start, upload_end)} seconds"
        )


class SystemsManagerOperator(object):
    def __init__(self, key: str = None, secret: str = None):
        """
        Initializes SystemsManagerOperator instance with given credentials. If no credentials are supplied, uses environment variables

        :param key: AWS access key id, default is None
        :param secret: AWS secret access key, default is None
        """
        self.client = boto3.client(
            "ssm",
            aws_access_key_id=key,
            aws_secret_access_key=secret,
        )

    def get_param(self, parameter_name: str):
        """
        Returns the value of requested parameter.

        :param parameter_name: name of the parameter requested
        :return:
        """
        parameter = self.client.get_parameter(Name=parameter_name, WithDecryption=True)
        return parameter["Parameter"]["Value"]



from google.cloud import bigquery


def connect(service_account_file: str):
    """
    Connects to Bigquery client with given service account file.

    :param service_account_file: relative location of service account json
    :return: a Client object instance required for API requests
    """
    return bigquery.Client.from_service_account_json(service_account_file)



import requests
from airflow_commons.internal.util.time_utils import datetime_add
from airflow_commons.logger import get_logger
from airflow_commons.internal.salesforce.constants import (
    GRANT_TYPE,
    AUTH_URL,
    TOKEN_PATH,
)
from datetime import datetime


class Connection:
    def __init__(
        self,
        client_id: str,
        client_secret: str,
        account_id: str,
        subdomain: str,
        scope: str = None,
    ):
        """
        Creates a connection instance to salesforce marketing cloud API

        :param client_id: Account client id string
        :param client_secret: Account client secret
        :param account_id: Account id
        :param subdomain: Account specific subdomain
        :param scope: Authorization scope
        """
        self.client_id = client_id
        self.client_secret = client_secret
        self.account_id = account_id
        self.auth_url = AUTH_URL.format(subdomain=subdomain)
        self.scope = scope
        self.logger = get_logger("Connection")
        self._authenticate()

    def _authenticate(self):
        """
        Authenticates to salesforce marketing cloud
        :return:
        """
        body = {
            "grant_type": GRANT_TYPE,
            "client_id": self.client_id,
            "client_secret": self.client_secret,
            "account_id": self.account_id,
        }
        if self.scope is not None:
            body["scope"] = self.scope
        now = datetime.now()
        auth_response = requests.post(url=(self.auth_url + TOKEN_PATH), json=body)
        try:
            auth_response.raise_for_status()
        except Exception as e:
            if auth_response:
                self.logger.error(
                    "An error occured during authentication " + auth_response.text
                )
            raise Exception(str(e)) from e
        auth_response_json = auth_response.json()
        self.access_token = auth_response_json["access_token"]
        self.token_expiration_period = auth_response_json["expires_in"] - 60
        self.soap_instance_url = auth_response_json["soap_instance_url"]
        self.rest_instance_url = auth_response_json["rest_instance_url"]
        self.token_expiration_at = datetime_add(
            now, seconds=self.token_expiration_period
        )

    def check_and_refresh_token(self):
        """
        Checks if current token is expired, if so re-authenticates itself.
        :return:
        """
        now = datetime.now()
        if now > self.token_expiration_at:
            return
        self._authenticate()



from google.cloud import bigquery


MEGABYTES_BILLED_DENOMINATOR = 2**20
AVAILABLE_COLUMN_NAMES = ["column_name", "table_name"]
SELECT_RETURN_TYPES = ["dict", "json", "dataframe", "pyarrow", "rowIterator"]
PRIMITIVE_DATA_TYPES = [
    "BOOL",
    "BYTES",
    "STRING",
    "DATE",
    "DATETIME",
    "TIME",
    "TIMESTAMP",
    "INT64",
    "FLOAT64",
    "NUMERIC",
    "BIGNUMERIC",
    "STRUCT",
]
JOB_PRIORITIES = ["INTERACTIVE", "BATCH"]
WRITE_DISPOSITIONS = ["WRITE_APPEND", "WRITE_TRUNCATE", "WRITE_EMPTY"]

JOB_CONFIGS = {
    "LoadJobConfig": bigquery.LoadJobConfig(),
    "QueryJobConfig": bigquery.QueryJobConfig(),
}



import logging
from time import gmtime

LOG_LEVEL_MAP = {"info": logging.INFO, "debug": logging.DEBUG, "error": logging.ERROR}


def get_logger(name: str, log_level: str = "info"):
    logging.Formatter.converter = gmtime
    logging.basicConfig(
        level=LOG_LEVEL_MAP[log_level],
        format="%(asctime)s | %(levelname)s | %(name)s | %(filename)s | %(funcName)s | %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
        force=True,
    )
    return logging.getLogger(name)



from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker


def get_session(
    username: str,
    password: str,
    host: str,
    db_name: str = None,
):
    """

    :param username: database username
    :param password: database password
    :param host: database host
    :param db_name: database name
    :return:
    """
    db_connection_str = "mysql+pymysql://{username}:{password}@{host}/{dbname}".format(
        username=username, password=password, host=host, dbname=db_name
    )

    engine = create_engine(db_connection_str)
    session = sessionmaker(engine)

    return session



from airflow_commons.internal.salesforce.constants import (
    ASYNC_PATH,
    DATA_EXTENSION_KEY_PATH,
    ASYNC_JOB_RESULT_PATH,
    ASYNC_JOB_STATUS_PATH,
    DATA_PATH,
    CUSTOM_OBJECT_DATA_PATH,
    FILTER_QUERY,
    SYNC_PATH,
    DATA_EVENT_KEY_PATH,
)
from airflow_commons.resources.glossary import QUESTION_MARK


def get_headers(access_token: str):
    """
    Builds a request header with given token.
    :param access_token: Active access token
    :return:
    """
    return {
        "Content-Type": "application/json",
        "Authorization": "Bearer " + access_token,
    }


def get_async_operation_url(base_url: str, key: str):
    """
    Builds an async operation url
    :param base_url:
    :param key:
    :return:
    """
    return base_url + ASYNC_PATH + DATA_EXTENSION_KEY_PATH.format(key=key)


def get_async_operation_result_url(base_url: str, request_id: str):
    """
    Builds an async operation result url
    :param base_url:
    :param request_id:
    :return:
    """
    return base_url + ASYNC_PATH + ASYNC_JOB_RESULT_PATH.format(request_id=request_id)


def get_async_operation_status_url(base_url: str, request_id: str):
    """
    Builds an async operation status url
    :param base_url:
    :param request_id:
    :return:
    """
    return base_url + ASYNC_PATH + ASYNC_JOB_STATUS_PATH.format(request_id=request_id)


def get_fetch_operation_url(base_url: str, key: str, filter: str = None):
    """
    Builds a get operation url
    :param base_url:
    :param key:
    :param filter: optional filter of search query
    :return:
    """
    url = base_url + DATA_PATH + CUSTOM_OBJECT_DATA_PATH.format(key=key)
    if filter:
        url = url + QUESTION_MARK + FILTER_QUERY.format(filter=filter)

    return url


def get_sync_operation_url(base_url: str, key: str):
    """
    Builds a sync operation url
    :param base_url:
    :param key:
    :return:
    """
    return base_url + SYNC_PATH + DATA_EVENT_KEY_PATH.format(key=key)
